import click
import sys
import time

import carica_dynamodb_tools.version
import carica_dynamodb_tools.version
from carica_dynamodb_tools.session import boto_session

WAIT_SLEEP_SECONDS = 5


@click.command()
@click.option('--region', '-r', help='AWS region name')
@click.option(
    '--wait',
    '-w',
    help='Wait until the export finishes',
    is_flag=True,
)
@click.argument('table')
@click.argument('bucket-name')
@click.version_option(version=carica_dynamodb_tools.version.__version__)
def cli(region: str, table: str, bucket_name: str, wait: bool):
    """
    Export a DynamoDB table to an S3 bucket at a point-in-time of "now".

    The export ARN is printed to stdout as soon as the export is created.
    You can feed it to carica-dynamodb-dump-s3-export to retrieve the dumped
    data when the export is COMPLETED.

    If you use the --wait flag, the export ARN is printed to stdout and the
    process waits until the export finishes.  If the export finishes with
    a COMPLETED status, the process's exit status is 0.  If the export
    finishes with a FAILED status, the process's exit status is 1.
    """
    session = boto_session(region_name=region)
    dynamodb_client = session.client('dynamodb')

    table_arn = dynamodb_client.describe_table(TableName=table)['Table']['TableArn']

    export_arn = dynamodb_client.export_table_to_point_in_time(
        TableArn=table_arn,
        ExportFormat='DYNAMODB_JSON',
        S3Bucket=bucket_name,
    )['ExportDescription']['ExportArn']

    print(export_arn)
    sys.stdout.flush()

    if wait:
        while True:
            desc = dynamodb_client.describe_export(ExportArn=export_arn)[
                'ExportDescription'
            ]
            status = desc['ExportStatus']
            if status == 'IN_PROGRESS':
                time.sleep(WAIT_SLEEP_SECONDS)
                continue
            elif status == 'COMPLETED':
                break
            else:
                print(f'ExportStatus is {status}', file=sys.stderr)
                sys.exit(1)


if __name__ == '__main__':
    cli()



import multiprocessing
import random
from multiprocessing.queues import Queue

import click
import orjson
import sys
import time
from botocore.exceptions import ClientError
from click import BadParameter

import carica_dynamodb_tools.version
import carica_dynamodb_tools.version

# DynamoDB API limit
from carica_dynamodb_tools.session import boto_session

BATCH_MAX_ITEMS = 25

# Initial retry delay in milliseconds
RETRY_DELAY_BASE_MS = 50

# Maximum retry delay in milliseconds
RETRY_DELAY_CAP_MS = 5000


def sleep_for_retry(attempt: int) -> None:
    """
    Sleep for an appropriate amount of time after a failed batch write.

    ``attempt`` should be 1 on the first call, 2 on the next call, etc.

    Implements "Full Jitter" from https://aws.amazon.com/blogs/architecture/exponential-backoff-and-jitter/
    """
    max_delay_ms = min(RETRY_DELAY_CAP_MS, RETRY_DELAY_BASE_MS * pow(2, attempt - 1))
    jittered_delay_ms = random.randint(0, max_delay_ms)
    jittered_delay_secs = jittered_delay_ms / 1000.0
    time.sleep(jittered_delay_secs)


def batch_worker(
    region: str,
    table: str,
    batch_q: Queue,
    print_lock: multiprocessing.Lock,
    success_total: multiprocessing.Value,
    failure_total: multiprocessing.Value,
    retry_total: multiprocessing.Value,
) -> None:
    """
    Multiprocessing worker for writing batches of items.

    Quits when it reads a ``None`` from the queue.
    """
    session = boto_session(region_name=region)
    client = session.client('dynamodb')
    for batch in iter(batch_q.get, None):
        request_items = [{'PutRequest': {'Item': item}} for item in batch]

        # Iterate until all the items have been written.  If BatchWriteItem
        # returns a throttling error response, that means none of the items
        # were written, and boto will retry the call automatically.  If any of
        # the items were successfully written, BatchWriteItem doesn't return an
        # error response.  Instead its response contains the unprocessed items
        # and we retry those in the next attempt (after a short delay).
        attempt = 1
        while request_items:
            num_unprocessed = 0
            num_processed = 0
            num_failed = 0

            try:
                resp = client.batch_write_item(RequestItems={table: request_items})
                unprocessed_items = resp.get('UnprocessedItems', {}).get(table, [])
                num_unprocessed = len(unprocessed_items)
                num_processed = len(request_items) - num_unprocessed
                # Put unprocessed items back in the list to be retried.
                request_items = unprocessed_items
            except ClientError as e:
                # boto3 automatically retries on throttling errors, so if we're
                # here, it's for something else (validation, internal error,
                # permissions, etc.) that retrying probably won't solve.
                with print_lock:
                    print(str(e), file=sys.stderr)
                num_failed = len(request_items)
                # Abandon them to prevent an infinite retry loop.
                request_items = []

            # Update the totals
            with success_total.get_lock():
                success_total.value += num_processed
            with retry_total.get_lock():
                retry_total.value += num_unprocessed
            with failure_total.get_lock():
                failure_total.value += num_failed

            if num_unprocessed:
                sleep_for_retry(attempt)
                attempt += 1


@click.command()
@click.option('--region', '-r', help='AWS region name')
@click.option(
    '--procs', '-p', help='Number of processes to use', default=4, show_default=True
)
@click.option(
    '--report/--no-report', '-r', help='Write a JSON report to stdout when finished'
)
@click.argument('table')
@click.version_option(version=carica_dynamodb_tools.version.__version__)
def cli(region: str, procs: str, report: bool, table: str):
    """
    Load items into a DynamoDB table from stdin, one JSON item per line.
    """
    num_procs = int(procs)
    if num_procs < 1:
        raise BadParameter('must be > 0', param_hint='procs')

    # Limiting the queue size puts backpressure on the producer.
    batch_q = multiprocessing.Queue(maxsize=num_procs * 10)
    print_lock = multiprocessing.Lock()
    success_total = multiprocessing.Value('i')
    failure_total = multiprocessing.Value('i')
    retry_total = multiprocessing.Value('i')
    proc_args = (
        region,
        table,
        batch_q,
        print_lock,
        success_total,
        failure_total,
        retry_total,
    )
    procs = [
        multiprocessing.Process(target=batch_worker, args=proc_args)
        for _ in range(num_procs)
    ]

    for p in procs:
        p.start()

    # Read items from stdin, batch them up, and send batches to workers.
    batch = []
    while True:
        line = sys.stdin.readline()

        # Squash empty lines (a newline char) into empty strings
        if line:
            line = line.strip()

        if line:
            batch.append(orjson.loads(line))

        # Send the batch to the worker when it's full or stdin is closed.
        # We don't worry about checking for total batch size here (in
        # serialized JSON bytes) because we assume DynamoDB would not have
        # supplied records that are too large to load during the dump process.
        if len(batch) == BATCH_MAX_ITEMS or not line:
            batch_q.put(batch)
            batch = []

        if not line:
            break

    for _ in procs:
        batch_q.put(None)

    for p in procs:
        p.join()

    if report:
        with success_total.get_lock(), failure_total.get_lock(), retry_total.get_lock():
            report = {
                'loaded': success_total.value,
                'failed': failure_total.value,
                'retried': retry_total.value,
            }
        sys.stdout.buffer.write(orjson.dumps(report))


if __name__ == '__main__':
    cli()



import click
import orjson
import sys

import carica_dynamodb_tools.version
import carica_dynamodb_tools.version
from carica_dynamodb_tools.session import boto_session
from carica_dynamodb_tools.utils import remove_protected_attrs


@click.command()
@click.option('--region', '-r', help='AWS region name')
@click.argument('table')
@click.version_option(version=carica_dynamodb_tools.version.__version__)
def cli(region, table):
    """
    Dump a DynamoDB table's items to stdout, one JSON item per line.

    Protected attributes (those starting with "aws:") are not included in output.
    """
    session = boto_session(region_name=region)
    client = session.client('dynamodb')
    paginator = client.get_paginator('scan')
    for page in paginator.paginate(TableName=table):
        for item in page['Items']:
            item = remove_protected_attrs(item)
            sys.stdout.buffer.write(orjson.dumps(item))
            sys.stdout.buffer.write(b'\n')


if __name__ == '__main__':
    cli()



import gzip
import multiprocessing
from multiprocessing import Queue
from typing import Tuple

import click
import orjson
import sys
from botocore.response import StreamingBody
from click import BadParameter

import carica_dynamodb_tools.version
import carica_dynamodb_tools.version
from carica_dynamodb_tools.session import boto_session
from carica_dynamodb_tools.utils import remove_protected_attrs


def get_export_data_items(
    region: str, export_arn: str
) -> Tuple[str, int, StreamingBody]:
    """
    :return: a tuple containing the bucket name, item count, and the response body
        to read the data files manifest JSON object
    """
    session = boto_session(region_name=region)
    dynamodb_client = session.client('dynamodb')
    s3_client = session.client('s3')

    desc = dynamodb_client.describe_export(ExportArn=export_arn)['ExportDescription']
    if desc['ExportFormat'] != 'DYNAMODB_JSON':
        print(f'ExportFormat is not DYNAMODB_JSON', file=sys.stderr)
        sys.exit(1)
    export_status = desc['ExportStatus']
    if export_status != 'COMPLETED':
        print(f'ExportStatus is {export_status}', file=sys.stderr)
        sys.exit(1)

    bucket = desc['S3Bucket']
    prefix = desc.get('S3Prefix', '')
    manifest_key = desc['ExportManifest']

    # Download the small export manifest JSON file
    resp = s3_client.get_object(Bucket=bucket, Key=f'{prefix}{manifest_key}')
    manifest = orjson.loads(resp['Body'].read())
    item_count = manifest['itemCount']
    manifest_files_key = manifest['manifestFilesS3Key']

    # Open the data items manifest JSON file, but return the response so the caller
    # can stream lines out of it.
    resp = s3_client.get_object(Bucket=bucket, Key=manifest_files_key)
    return bucket, item_count, resp['Body']


def batch_worker(
    region: str,
    bucket: str,
    item_q: Queue,
    print_lock: multiprocessing.Lock,
    item_total: multiprocessing.Value,
    error_total: multiprocessing.Value,
) -> None:
    """
    Multiprocessing worker for dumping JSONL archives in S3.

    Quits when it reads a ``None`` from the queue.
    """
    session = boto_session(region_name=region)
    s3_client = session.client('s3')
    for manifest_item in iter(item_q.get, None):
        # The manifest item is the contents of one manifestFilesS3Key file.  It looks like:
        # {
        #   'dataFileS3Key': 'AWSDynamoDB/01680958677849-381aef7c/data/s3rcacg63a6lfieybvp7dw357y.json.gz',
        #   'etag': 'ba00d841bd1eec340400e8d62c778aa3-1',
        #   'itemCount': 5344,
        #   'md5Checksum': 'R66Q93z6mjqdBW/mkgj0/A==',
        # }
        key = manifest_item['dataFileS3Key']
        etag = manifest_item['etag']
        try:
            resp = s3_client.get_object(Bucket=bucket, Key=key, IfMatch=etag)
            with gzip.open(resp['Body'], 'rt') as data_item_lines:
                for data_item_line in data_item_lines:
                    # Remove the wrapping "Item" property at the top level
                    item = orjson.loads(data_item_line)['Item']
                    item = remove_protected_attrs(item)
                    item_json = orjson.dumps(item)
                    with print_lock:
                        # Appending the newline and calling write one time, instead of
                        # writing the newline with a second call, prevents an apparent
                        # race where the underlying output doesn't stay consistent when
                        # multiple threads write to it quickly, even with print_lock.
                        # The effect of that condition is extra or misplaced newlines
                        # in the output, which we don't want.
                        sys.stdout.buffer.write(item_json + b'\n')
                    with item_total.get_lock():
                        item_total.value += 1
        except Exception as e:
            with error_total.get_lock():
                error_total.value += 1
            with print_lock:
                print(
                    f'Error getting s3://{bucket}/{key.lstrip("/")} etag={etag}: {e}',
                    file=sys.stderr,
                )


@click.command()
@click.option('--region', '-r', help='AWS region name')
@click.option(
    '--procs', '-p', help='Number of processes to use', default=4, show_default=True
)
@click.argument('export-arn')
@click.version_option(version=carica_dynamodb_tools.version.__version__)
def cli(region: str, procs: int, export_arn: str):
    """
    Dump all items in a JSON-format S3 export of a DynamoDB table to stdout,
    one JSON item per line.

    When you export a DynamoDB table to S3 in JSON format, DynamoDB writes
    JSONL objects to S3, but those objects contain a serialized format of
    the DynamoDB item that places each item's data inside a top-level "Item"
    attribute at the root of its JSONL line.  The output of this command
    removes that nesting technique, returning each item's serialized data
    at the root of its JSONL line.  This makes the output compatible with
    the "dump" command.

    Protected attributes (those starting with "aws:") are not included in output.
    """
    num_procs = int(procs)
    if num_procs < 1:
        raise BadParameter('must be > 0', param_hint='procs')

    bucket, item_count, data_manifest_response = get_export_data_items(
        region, export_arn
    )

    # Limiting the queue size puts backpressure on the producer.
    manifest_item_q = multiprocessing.Queue(maxsize=num_procs * 10)
    item_total = multiprocessing.Value('i')
    error_total = multiprocessing.Value('i')

    print_lock = multiprocessing.Lock()
    proc_args = (
        region,
        bucket,
        manifest_item_q,
        print_lock,
        item_total,
        error_total,
    )
    procs = [
        multiprocessing.Process(target=batch_worker, args=proc_args)
        for _ in range(num_procs)
    ]

    for p in procs:
        p.start()

    # Read manifest items, putting one decoded item into the queue at a time.
    # Put blocks when the queue is full.
    for line in data_manifest_response.iter_lines():
        manifest_item_q.put(orjson.loads(line))

    for _ in procs:
        manifest_item_q.put(None)

    for p in procs:
        p.join()

    error = False
    if item_total.value != item_count:
        print(
            f'Expected {item_count} items; {item_total.value} items in backup',
            file=sys.stderr,
        )
        error = True
    if error_total.value > 0:
        print(f'{error_total.value} errors getting item data', file=sys.stderr)
        error = True
    if error:
        sys.exit(1)


if __name__ == '__main__':
    cli()



#!/usr/bin/env python3
"""A setuptools based setup module.
See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
from codecs import open
from os import path

from setuptools import setup, find_packages

import carica_dynamodb_tools.version

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='carica_dynamodb_tools',
    version=carica_dynamodb_tools.version.__version__,
    description='Tools to manage DynamoDB tables',
    long_description=long_description,
    url='https://github.com/caricalabs/carica-dynamodb-tools',
    author='Carica Labs, LLC',
    author_email='info@caricalabs.com',
    license='APL 2.0',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3.7',
    ],
    keywords='dynamodb backup restore archive dump load',
    packages=find_packages(exclude=['contrib', 'docs', 'tests']),
    install_requires=[
        'boto3>=1.9.99',
        'click~=8.0',
        'orjson~=3.9.2',
    ],
    extras_require={
        'dev': ['check-manifest'],
        'test': [],
    },
    package_data={},
    entry_points={
        'console_scripts': [
            'carica-dynamodb-dump=carica_dynamodb_tools.dump:cli',
            'carica-dynamodb-dump-s3-export=carica_dynamodb_tools.dump_s3_export:cli',
            'carica-dynamodb-load=carica_dynamodb_tools.load:cli',
            'carica-dynamodb-s3-export=carica_dynamodb_tools.s3_export:cli',
        ],
    },
)



def remove_protected_attrs(item: dict) -> dict:
    """
    Remove protected (AWS-only) attributes from a DynamoDB item.
    """
    attrs = [attr for attr in item.keys() if attr.startswith('aws:')]
    for attr in attrs:
        del item[attr]
    return item



import boto3.session
import botocore.config
import botocore.session


def boto_session(*args, **kwargs) -> boto3.session.Session:
    """
    Get a boto3 session configured for more retries.
    """
    botocore_session = botocore.session.Session()
    botocore_session_config = botocore.config.Config(
        retries={
            'mode': 'standard',
            'max_attempts': 20,
        },
        max_pool_connections=100,
    )
    botocore_session.set_default_client_config(botocore_session_config)
    return boto3.session.Session(*args, **kwargs, botocore_session=botocore_session)
